import os
import torch.nn as nn
from torch import optim
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import DataLoader
from utils import *
from model.GCNN3d import GCNN3d
from data.dataloader import *
import scipy.io as scio
from data.data_utils import kappa_func
import glob

if __name__ == '__main__':
    # Select a data set
    # data = Houston2018_data(norm="stand")
    # data = XiongAn_data(norm="stand")
    data = PaviaU_data(norm="stand")
    # data = Indian_Pines_data(norm="stand")
    # data = GF5_data(norm="stand")

    class_num = data.getClassNum()

    # GPU parallel
    os.environ['CUDA_VISIBLE_DEVICES'] = "0"
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    use_full_index = 0

    if use_full_index == 0:

        epoch = 120
        test_step = 120
        repeat_time = 3

        for sam_num in range(5, 7):
            rate = np.zeros((class_num + 2, repeat_time + 1))
            kappa = np.zeros((repeat_time + 1, 1))

            for ite in range(repeat_time):
                # Load data
                train_data = HLDataSet(data=data, sam_num=sam_num, ite=ite, train=True, r=2)
                train_loader = DataLoader(dataset=train_data, batch_size=1024, num_workers=4,
                                          pin_memory=True, shuffle=True, drop_last=False)

                test_data = HLDataSet(data=data, sam_num=sam_num, ite=ite, train=False, r=2)
                test_loader = DataLoader(dataset=test_data, batch_size=1024, num_workers=4,
                                         pin_memory=True, shuffle=False, drop_last=False)

                # Load model
                model = GCNN3d()
                model = model.to(device)

                # Training setting
                # optimizer = optim.Adadelta(model.parameters(), lr=0.001, rho=0.9, eps=1e-06, weight_decay=3e-05)
                optimizer = optim.Adam(model.parameters(), lr=0.001, weight_decay=0.001)
                # optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.4, weight_decay=3e-05)
                scheduler = StepLR(optimizer, step_size=20, gamma=0.5)  # 0.5

                loss_func = nn.CrossEntropyLoss().to(device)

                best_test_acc = 0
                loss_record = []
                # train_loss_record = []
                # train_acc_record = []
                # test_acc_record = []

                for e in range(epoch):
                    # Train
                    model.train()
                    train_acc = []
                    for batch_idx, (cube, label) in enumerate(train_loader):
                        cube = torch.squeeze(cube, 1)
                        cube, label = cube.to(device), label.to(device)
                        optimizer.zero_grad()
                        output = model(cube)
                        train_acc.append(cal_accuracy(output, label))
                        loss = loss_func(output, label)
                        loss.backward()
                        optimizer.step()
                        loss_record.append(loss.item())
                    print('sample number: {:d}\titeration: {:d}\tepoch: {:d}\ttrain_loss: {:.4f}\ttrain_acc: {:.2f}%'
                          .format(sam_num, ite, e + 1, np.mean(loss_record), np.mean(train_acc)))
                    # train_loss_record.append(np.mean(loss_record))
                    # train_acc_record.append(np.mean(train_acc))
                    scheduler.step()

                    # Estimation
                    if (e + 1) % test_step == 0:
                        mix_matrix = np.zeros((class_num, class_num))
                        predict_map = np.array([])

                        model.eval()
                        with torch.no_grad():
                            for batch_idx, batch in enumerate(test_loader):
                                cube, label = batch[0], batch[1]
                                cube = torch.squeeze(cube, 1)
                                cube, label = cube.to(device), label.to(device)
                                output = model(cube)
                                predicted = output.argmax(1)

                                predict_map = np.append(predict_map, np.array(predicted.to(torch.device("cpu"))))

                                for i, r_c in enumerate(zip(predicted, label)):
                                    r, c = r_c
                                    mix_matrix[r, c] += 1

                        class_sequence = list(range(class_num))

                        # Overall accuracy
                        correct = float(np.sum(mix_matrix[class_sequence, class_sequence]))
                        total = float(np.sum(mix_matrix))
                        rate[class_num, ite] = correct / (total + 1e-6)
                        test_acc = rate[class_num, ite]

                        # Class accuracy
                        cls_accuracy = 0.0
                        for cls in range(class_num):
                            correct = float(mix_matrix[cls, cls])
                            total = float(np.sum(mix_matrix[:, cls]))
                            rate[cls, ite] = correct / (total + 1e-6)
                            cls_accuracy += correct / (total + 1e-6)

                        # Average accuracy
                        rate[class_num + 1, ite] = cls_accuracy / class_num

                        # Kappa
                        kappa[ite, 0] = kappa_func(mix_matrix)

                        # test_acc_record.append(np.mean(test_acc))
                        print('\nsample number: {:d}\titeration: {:d}\tavg_test_acc: {:.2f}%\n'
                              .format(sam_num, ite, 100 * test_acc))

                        # Record best test accuracy
                        is_best = test_acc > best_test_acc
                        best_test_acc = max(test_acc, best_test_acc)

                        # Save checkpoint
                        save_checkpoint(
                            {'e': e,
                             'state_dict': model.state_dict(),
                             'best_test_acc': best_test_acc,
                             'optimizer': optimizer.state_dict()},
                            is_best,
                            filename='/home/kinfai/Code/3DGCNN_34/checkpoint/{}_{}_{}_{}.pth.tar'
                                .format(model.model_name, data.dataName, sam_num, ite))

                        # Save predict map
                        if is_best:
                            path = '/home/kinfai/Code/3DGCNN_34/result/predict_map_{}_{}_{}_{}_*.mat' \
                                .format(data.dataName, model.model_name, sam_num, ite)
                            if len(glob.glob(path)) > 0:
                                os.remove(glob.glob(path)[0])
                            path = '/home/kinfai/Code/3DGCNN_34/result/predict_map_{}_{}_{}_{}_{}.mat' \
                                .format(data.dataName, model.model_name, sam_num, ite, int(best_test_acc * 1000 + 0.5))
                            scio.savemat(path, {'predict_map': predict_map})

            # Save test result
            rate[:, repeat_time] = np.mean(rate[:, 0:repeat_time], axis=1)
            kappa[repeat_time, 0] = np.mean(kappa)

            result = {'rate': rate, 'kappa': kappa}
            path = '/home/kinfai/Code/3DGCNN_34/result/{}_{}_{}_{}.mat' \
                .format(data.dataName, model.model_name, sam_num, int(rate[class_num, repeat_time] * 1000 + 0.5))
            sio.savemat(path, {model.model_name: result})

    else:
        # Save full index predict map ==================================================================================
        sam_num = 4
        ite = 0

        # Load data
        test_data_full_index = HLDataSet(data=data, sam_num=sam_num, ite=ite, train=False, useFullIndex=True)
        test_loader_full_index = DataLoader(dataset=test_data_full_index, batch_size=1024, num_workers=4,
                                            pin_memory=True, shuffle=False, drop_last=False)

        predict_map_full_index = np.array([])

        # Load model
        model = GCNN3d()
        model = model.to(device)
        checkpoint = torch.load('/home/kinfai/Code/3DGCNN_34/checkpoint/{}_{}_{}_{}.pth.tar'
                                .format(model.model_name, data.dataName, sam_num, ite))
        model.load_state_dict(checkpoint['state_dict'])

        model.eval()
        with torch.no_grad():
            for batch_idx, batch in enumerate(test_loader_full_index):
                cube = torch.squeeze(batch[0], 1).to(device)
                output = model(cube)
                predicted = output.argmax(1)
                predict_map_full_index = np.append(predict_map_full_index,
                                                   np.array(predicted.to(torch.device("cpu"))) + 1)

        path = '/home/kinfai/Code/3DGCNN_34/result/predict_map_fullindex_{}_{}_{}.mat' \
            .format(data.dataName, model.model_name, sam_num)
        scio.savemat(path, {'predict_map': predict_map_full_index})