import torch
import torch.nn as nn
import torch.nn.init as init
import GConv3d


class GCNN3d(nn.Module):
    def __init__(self):
        super(GCNN3d, self).__init__()
        self.model_name = 'GCNN3d'
        self.gabor_channel = 13
        self.band_num = 185  # pu_103/gf5_280/hou_48/xa_256/ip_185
        self.class_num = 20

        self.GConv1_1 = GConv3d(self.band_num // self.gabor_channel, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=1, expand=True)
        self.GConv1_2 = GConv3d(64, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=1, bias=False)

        self.GConv2_1 = GConv3d(self.band_num // self.gabor_channel, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=2, expand=True)
        self.GConv2_2 = GConv3d(64, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=2, bias=False)

        self.GConv3_1 = GConv3d(self.band_num // self.gabor_channel, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=3, expand=True)
        self.GConv3_2 = GConv3d(64, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=3, bias=False)

        self.GConv4_1 = GConv3d(self.band_num // self.gabor_channel, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=4, expand=True)
        self.GConv4_2 = GConv3d(64, 64, 3, padding=0, stride=1, M=self.gabor_channel, nScale=4, bias=False)

        self.fc1 = nn.Linear(256, 128)
        self.fc2 = nn.Linear(128, self.class_num)

        self.bn = nn.BatchNorm2d(64 * self.gabor_channel)
        self.relu = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout(p=0.5)

        for m in self.modules():
            if isinstance(m, GConv3d):
                init.orthogonal_(m.weight)
                # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)

    def forward(self, x):
        x_1 = self.relu(self.bn(self.GConv1_1(x)))
        x_1 = self.relu(self.bn(self.GConv1_2(x_1)))
        x_1 = x_1.view(-1, 64, self.gabor_channel)

        x_2 = self.relu(self.bn(self.GConv2_1(x)))
        x_2 = self.relu(self.bn(self.GConv2_1(x_2)))
        x_2 = x_2.view(-1, 64, self.gabor_channel)

        x_3 = self.relu(self.bn(self.GConv3_1(x)))
        x_3 = self.relu(self.bn(self.GConv3_2(x_3)))
        x_3 = x_3.view(-1, 64, self.gabor_channel)

        x_4 = self.relu(self.bn(self.GConv4_1(x)))
        x_4 = self.relu(self.bn(self.GConv4_2(x_4)))
        x_4 = x_4.view(-1, 64, self.gabor_channel)

        x = torch.cat([x_1, x_2, x_3, x_4], 1)
        x = torch.max(x, 2)[0]
        # x = torch.mean(x, 2)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.fc2(x)

        return x

